import shutil
from napkinxc.datasets import load_dataset
from napkinxc.models import PLT

from conf import *
MODEL_PATH = get_model_path(__file__)


def test_set_get_tree_structure():
    X, Y = load_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)
    plt = PLT(MODEL_PATH)
    plt.build_tree(X, Y)
    tree_structure = plt.get_tree_structure()
    plt.set_tree_structure(tree_structure)
    tree_structure2 = plt.get_tree_structure()
    assert len(set(tree_structure) - set(tree_structure2)) == 0

    nodes_to_update = plt.get_nodes_to_update(Y)
    assert len(nodes_to_update) == X.shape[0]

    nodes_updates = plt.get_nodes_updates(Y)
    assert len(nodes_updates) == len(tree_structure)

    plt.fit(X, Y)
    tree_structure3 = plt.get_tree_structure()
    assert len(set(tree_structure) - set(tree_structure3)) == 0

    shutil.rmtree(MODEL_PATH, ignore_errors=True)


def test_build_tree_structure_reproducibility():
    X, Y = load_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)
    plt = PLT(MODEL_PATH + "-1", seed=1993)
    plt.build_tree(X, Y)
    tree_structure = plt.get_tree_structure()

    plt2 = PLT(MODEL_PATH + "-2", seed=1993)
    plt2.build_tree(X, Y)
    tree_structure2 = plt2.get_tree_structure()

    assert len(set(tree_structure) - set(tree_structure2)) == 0

    shutil.rmtree(MODEL_PATH + "-1", ignore_errors=True)
    shutil.rmtree(MODEL_PATH + "-2", ignore_errors=True)



# Copyright (c) 2020-2022 by Marek Wydmuch
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import gdown
import gzip
import json
import re
import requests
import warnings
import zipfile
from sys import stdout
from os import makedirs, path, remove
import numpy as np
from scipy.sparse import csr_matrix

try:
    from ._napkinxc import _load_libsvm_file_labels_list, _load_libsvm_file_labels_csr_matrix
except ImportError:
    warnings.warn("Couldn't import napkinXC cpp module, some functions may fail.")


# List of all available datasets
DATASETS = {
    # Small datasets, for testing
    'yeast': {
        'name': 'yeast',
        'formats': ['bow'],
        'subsets': ['train', 'test'],
        'bow': {
            'url': 'https://drive.google.com/uc?export=download&id=1qcU-jxF7VmHBvwTn0-6MB4FW4ItD-bIp',  # Marek Wydmuch's upload
            'dir': 'yeast',
            'train': 'yeast_train.txt',
            'test': 'yeast_test.txt',
            'file_format': 'libsvm',
        }
    },
    'mediamill': {
        'name': 'mediamill',
        'formats': ['bow'],
        'subsets': ['train', 'test'],
        'bow': {
            'url': 'https://drive.google.com/uc?export=download&id=1YDlU9GTYSoydU1UYTNT8BpPkRw1BCnFw',  # Marek Wydmuch's upload
            'dir': 'mediamill',
            'train': 'mediamill_train.txt',
            'test': 'mediamill_test.txt',
            'file_format': 'libsvm',
        }
    },

    # XMLC repo datasets
    'rcv1-2k': {
        'name': 'RCV1-2K',
        'formats': ['bow'],
        'subsets': ['train', 'test'],
        'bow': {
            'url': 'https://drive.google.com/uc?export=download&id=1pdken_Di3PLcLlRbhHF4XIWp-CEiv9oQ',  # Marek Wydmuch's upload
            'dir': 'RCV1-x',
            'train': 'rcv1x_train.txt',
            'test': 'rcv1x_test.txt',
            'file_format': 'libsvm',
        }
    },
    'eurlex-4k': {
        'name': 'EURLex-4K',
        'formats': ['bow'],
        'subsets': ['train', 'test'],
        'bow': {
            # 'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGU0VTR1pCejFpWjg',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1xWW1UykQBTD2IorVn7V8XubJatQDScol',  # Marek Wydmuch's upload
            'dir': 'Eurlex',
            'train': 'eurlex_train.txt',
            'test': 'eurlex_test.txt',
            'file_format': 'libsvm',
        }
    },
    'eurlex-4.3k': {
        'name': 'EURLex-4.3K',
        'formats': ['bow'],
        'subsets': ['train', 'test', 'validation'],
        'bow': {
            # 'url': 'https://drive.google.com/uc?export=download&id=1b3mWgaKIAmc9Ae3E0QrokiIFA9Qj1K9r',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1UZ5joZdU1axWdCVSECakmEA8XQJ2gVuQ',  # Marek Wydmuch's upload
            'dir': 'EURLex-4.3K',
            'train': 'train.txt',
            'test': 'test.txt',
            'validation': 'validation.txt',
            'file_format': 'libsvm',
        }
    },
    'amazoncat-13k': {
        'name': 'AmazonCat-13K',
        'formats': ['bow-v1', 'bow', 'raw'],
        'subsets': ['train', 'test'],
        'bow-v1': {
            # 'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGa2tMbVJGdDNSMGc',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1WvBWF9hH5ZlFcpCRhp8wLVGwdeGZXKD2',  # Marek Wydmuch's upload
            'dir': 'AmazonCat',
            'train': 'amazonCat_train.txt',
            'test': 'amazonCat_test.txt',
            'file_format': 'libsvm',
        },
        'bow': {  # v2
            'url': 'https://drive.google.com/uc?export=download&id=17EFQtnRswEv0XyPng2EOy5IeWmeZbe0a',  # Marek Wydmuch's upload
            'dir': 'AmazonCat-13K.bow',
            'train': 'train.txt',
            'test': 'test.txt',
            'file_format': 'libsvm',
        },
        'raw': {
            # 'url': 'https://drive.google.com/uc?export=download&id=17rVRDarPwlMpb3l5zof9h34FlwbpTu4l',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1iOsEOYL-fY-GOt0CNfgW9YO2oFDTb6X4',  # Marek Wydmuch's upload
            'dir': 'AmazonCat-13K.raw',
            'train': 'trn.json.gz',
            'test': 'tst.json.gz',
            'file_format': 'jsonlines',
            'features_fields': ['title', 'content'],
            'labels_field': 'target_ind'
        }
    },
    'amazoncat-14k': {
        'name': 'AmazonCat-14K',
        'formats': ['bow-v1', 'bow', 'raw'],
        'subsets': ['train', 'test'],
        'bow-v1': {
            # 'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGaDFqU2E5U0dxS00',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=19IhTX1_a4U5I2Q56VUkzq2jFBpwm2-n-',  # Marek Wydmuch's upload
            'dir': 'AmazonCat-14K',
            'train': 'amazonCat-14K_train.txt',
            'test': 'amazonCat-14K_test.txt',
            'file_format': 'libsvm',
        },
        'bow': {  # v2
            'url': 'https://drive.google.com/uc?export=download&id=1VAxPvByB-SX3_MefdfW1eKZYTfz0vMcR',  # Marek Wydmuch's upload
            'dir': 'AmazonCat-14K.bow',
            'train': 'train.txt',
            'test': 'test.txt',
            'file_format': 'libsvm',
        },
        'raw': {
            # 'url': 'https://drive.google.com/uc?export=download&id=1vy1N-lDdDfuoo0CNwFE11hb3INCpJHFx',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1CTu0ZKhpaWuzIoOheDrccG_skXhyfzLm',  # Marek Wydmuch's upload
            'dir': 'AmazonCat-14K.raw',
            'train': 'trn.json.gz',
            'test': 'tst.json.gz',
            'file_format': 'jsonlines',
            'features_fields': ['title', 'content'],
            'labels_field': 'target_ind'
        }
    },
    'wiki10-31k': {
        'name': 'Wiki10-31K',
        'formats': ['bow-v1', 'bow'],
        'subsets': ['train', 'test'],
        'bow-v1': {
            # 'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGaDdOeGliWF9EOTA',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1uV-p682ggXQQTiKZyK5M6B0xJJk8_lJx',  # Marek Wydmuch's upload
            'dir': 'Wiki10',
            'train': 'wiki10_train.txt',
            'test': 'wiki10_test.txt',
            'file_format': 'libsvm',
        },
        'bow': {  # v2
            'url': 'https://drive.google.com/uc?export=download&id=1WbxhKvvBSynX5_Sf6OaPReTMs8K6bZ8u',  # Marek Wydmuch's upload
            'dir': 'Wiki10.bow',
            'train': 'train.txt',
            'test': 'test.txt',
            'file_format': 'libsvm',
        }
    },
    'deliciouslarge-200k': {
        'name': 'DeliciousLarge-200K',
        'formats': ['bow'],
        'subsets': ['train', 'test'],
        'bow': {
            # 'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGR3lBWWYyVlhDLWM',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=18Tb595TuGyFW--nEYGfduwS5WvSbX-sX',  # Marek Wydmuch's upload
            'dir': 'DeliciousLarge',
            'train': 'deliciousLarge_train.txt',
            'test': 'deliciousLarge_test.txt',
            'file_format': 'libsvm',
        }
    },
    'wikilshtc-325k': {
        'name': 'WikiLSHTC-325K',
        'formats': ['bow'],
        'subsets': ['train', 'test'],
        'bow': {
            # 'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGSHE1SWx4TVRva3c',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1Y5xEJu-j3M7voyFV5Ouq9Wl3vyd9fus2',  # Marek Wydmuch's upload
            'dir': 'WikiLSHTC',
            'train': 'wikiLSHTC_train.txt',
            'test': 'wikiLSHTC_test.txt',
            'file_format': 'libsvm',
        }
    },
    'wikiseealsotitles-350k': {
        'name': 'WikiSeeAlsoTitles-350K',
        'formats': ['bow', 'raw'],
        'subsets': ['train', 'test'],
        'bow': {
            # 'url': 'https://drive.google.com/uc?export=download&id=1bHtiLVF5EFsVL3qyU7y5e3M-fYHvXsG9',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1icZIEU2TA5k86mTweciSa3mBLFD3fbIP',  # Marek Wydmuch's upload
            # 'dir': 'WikiSeeAlsoTitles-350K',  # (XMLC version)
            'dir': 'WikiSeeAlsoTitles-350K.bow',
            'train': {'X': 'trn_X_Xf.txt', 'Y': 'trn_X_Y.txt'},
            'test': {'X': 'tst_X_Xf.txt', 'Y': 'tst_X_Y.txt'},
            'file_format': 'XY_sparse',
        },
        'raw': {
            # 'url': 'https://drive.google.com/uc?export=download&id=1sxPHzlnotUKjbtVRe0GuXGfSI7YhBSdc',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=13Y4cvJ6zKDUpIdBVsaBHp1rEWiv5tyQy',  # Marek Wydmuch's upload
            # 'dir': 'WikiSeeAlsoTItles-350K',  # (XMLC version)
            'dir': 'WikiSeeAlsoTitles-350K.raw',
            'train': 'trn.json.gz', # Type is there on purpose
            'test': 'tst.json.gz',
            'file_format': 'jsonlines',
            'features_fields': ['title', 'content'],
            'labels_field': 'target_ind'
        }
    },
    'wikititles-500k': {
        'name': 'WikiTitles-500K',
        'formats': ['bow', 'raw'],
        'subsets': ['train', 'test'],
        'bow': {
            # 'url': 'https://drive.google.com/uc?export=download&id=11U4qDWKvsR6pCzLvY3APckx-R_ihyMih',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1RKKOw0yl2RxrWYbb-S508aOwKdCjY_ih',  # Marek Wydmuch's upload
            'dir': 'WikiTitles-500K',
            'train': {'X': 'trn_X_Xf.txt', 'Y': 'trn_X_Y.txt'},
            'test': {'X': 'tst_X_Xf.txt', 'Y': 'tst_X_Y.txt'},
            'file_format': 'XY_sparse',
        },
        'raw': {
            'url': 'https://drive.google.com/uc?export=download&id=1YStqoa6_5Qxd9FpExTNt-_tcXUhYXUFM',  # Marek Wydmuch's reupload
            'dir': 'Wikipedia-500K.raw',
            'train': 'trn.raw.json.gz',
            'test': 'tst.raw.json.gz',
            'file_format': 'jsonlines',
            'features_fields': ['title'],
            'labels_field': 'target_ind'
        }
    },
    'wikipedialarge-500k': {
        'name': 'WikipediaLarge-500K',
        'formats': ['bow', 'raw'],
        'subsets': ['train', 'test'],
        'bow': {
            # 'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGRmEzVDVkNjBMR3c',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1LkGdINF5coOfkm6J2xys3V1NMW7zz8U1',  # Marek Wydmuch's upload
            'dir': 'WikipediaLarge-500K',
            'train': 'WikipediaLarge-500K_train.txt',
            'test': 'WikipediaLarge-500K_test.txt',
            'file_format': 'libsvm',
        },
        'raw': {
            'url': 'https://drive.google.com/uc?export=download&id=1YStqoa6_5Qxd9FpExTNt-_tcXUhYXUFM',  # Marek Wydmuch's reupload
            'dir': 'Wikipedia-500K.raw',
            'train': 'trn.raw.json.gz',
            'test': 'tst.raw.json.gz',
            'file_format': 'jsonlines',
            'features_fields': ['title', 'content'],
            'labels_field': 'target_ind'
        }
    },
    'amazontitles-670k': {
        'name': 'AmazonTitles-670K',
        'formats': ['bow', 'raw'],
        'subsets': ['train', 'test'],
        'bow': {
            # 'url': 'https://drive.google.com/uc?export=download&id=1OKnaLu4SDMOQ69rHdwF8ExKkeF2SZw7z',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1TWF2_tPGdezuljGEHbp63pe7eQUq7wK6',  # Marek Wydmuch's upload
            'dir': 'AmazonTitles-670K',
            'train': {'X': 'trn_X_Xf.txt', 'Y': 'trn_X_Y.txt'},
            'test': {'X': 'tst_X_Xf.txt', 'Y': 'tst_X_Y.txt'},
            'file_format': 'XY_sparse',
        },
        'raw': {
            # 'url': 'https://drive.google.com/uc?export=download&id=1FPqD8Wns7NXTSYDAcK4ZsqUUGABLyZMn',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1f1w0cGc-jJoPxJRY7v7BvKJuelspwOZV',  # Marek Wydmuch's upload
            'dir': 'AmazonTitles-670K',
            'train': 'trn.json.gz',
            'test': 'tst.json.gz',
            'file_format': 'jsonlines',
            'features_fields': ['title'],
            'labels_field': 'target_ind'
        }
    },
    'amazon-670k': {
        'name': 'Amazon-670K',
        'formats': ['bow-v1', 'bow', 'raw'],
        'subsets': ['train', 'test'],
        'bow-v1': {
            # 'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGdUJwRzltS1dvUVk',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1PV-wbKVv6Ng1K1XM1USqNxFhaQo2qfk6',  # Marek Wydmuch's upload
            'dir': 'Amazon',
            'train': 'amazon_train.txt',
            'test': 'amazon_test.txt',
            'file_format': 'libsvm',
        },
        'bow': {
            # 'url': 'https://drive.google.com/uc?export=download&id=1TLaXCNB_IDtLhk4ycOnyud0PswWAW6hR',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1d5EMgVxjjDBGjw7M09gaygx5wSt4-AqQ',  # Marek Wydmuch's upload
            # 'dir': 'Amazon670K.bow',  # (XMLC version)
            'dir': 'Amazon-670K.bow',
            'train': 'train.txt',
            'test': 'test.txt',
            'file_format': 'libsvm',
        },
        'raw': {
            # 'url': 'https://drive.google.com/uc?export=download&id=16FIzX3TnlsqbrwSJJ2gDih69laezfZWR',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1WAzs7qfLY8ZoluHbY55WyakhibW27kS2',  # Marek Wydmuch's upload
            'dir': 'Amazon-670K.raw',
            'train': 'trn.raw.json.gz',
            'test': 'tst.raw.json.gz',
            'file_format': 'jsonlines',
            'features_fields': ['title', 'content'],
            'labels_field': 'target_ind'
        }
    },
    'amazontitles-3m': {
        'name': 'AmazonTitles-3M',
        'formats': ['bow', 'raw'],
        'subsets': ['train', 'test'],
        'bow': {
            # 'url': 'https://drive.google.com/uc?export=download&id=1PGzippnzIcgVNYZ8qQKVb0GNiARjQxvV',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=15MpbBO7MTRLc1okzwSiPkQ9ZHgyKMzzm',  # Marek Wydmuch's upload
            'dir': 'AmazonTitles-3M',
            'train': {'X': 'trn_X_Xf.txt', 'Y': 'trn_X_Y.txt'},
            'test': {'X': 'tst_X_Xf.txt', 'Y': 'tst_X_Y.txt'},
            'file_format': 'XY_sparse',
        },
        'raw': {
            # 'url': 'https://drive.google.com/uc?export=download&id=1m0MMApC0vPpjEfI35SAaBGqKDsYypVXs',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=18Rpy2WWJL8AlGoipT3DSWehzu54QLH_7',  # Marek Wydmuch's upload
            'dir': 'AmazonTitles-3M',
            'train': 'trn.json.gz',
            'test': 'tst.json.gz',
            'file_format': 'jsonlines',
            'features_fields': ['title'],
            'labels_field': 'target_ind'
        }
    },
    'amazon-3m': {
        'name': 'Amazon-3M',
        'formats': ['bow-v1', 'bow', 'raw'],
        'subsets': ['train', 'test'],
        'bow-v1': {
            # 'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGUEd4eTRxaWl3YkE',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1ork7yeAcliD9JQiRdEx4QknhMFHjVrku',  # Marek Wydmuch's upload
            'dir': 'Amazon-3M',
            'train': 'amazon-3M_train.txt',
            'test': 'amazon-3M_test.txt',
            'file_format': 'libsvm',
        },
        'bow': {
            'url': 'https://drive.google.com/uc?export=download&id=1tepPa5oOSp7jmih7LTEemHOL9KwdBtuB',  # Marek Wydmuch's upload
            'dir': 'Amazon-3M.bow',
            'train': 'train.txt',
            'test': 'test.txt',
            'file_format': 'libsvm',
        },
        'raw': {
            # 'url': 'https://drive.google.com/uc?export=download&id=1gsabsx8KR2N9jJz16jTcA0QASXsNuKnN',  # XMLC repo url (stopped working)
            'url': 'https://drive.google.com/uc?export=download&id=1oDW_-c9dW6Dwj0ja_RVoLRNp87JKw1gh',  # Marek Wydmuch's upload
            'dir': 'Amazon-3M.raw',
            'train': 'trn.json.gz',
            'test': 'tst.json.gz',
            'file_format': 'jsonlines',
            'features_fields': ['title', 'content'],
            'labels_field': 'target_ind'
        }
    },
    'lf-amazontitles-131k': {
        'name': 'LF-AmazonTitles-131K',
        'formats': ['bow'],
        'subsets': ['train', 'test'],
        'bow': {
            'url': 'https://drive.google.com/uc?export=download&id=1VlfcdJKJA99223fLEawRmrXhXpwjwJKn',  # XMLC repo url
            'dir': 'LF-AmazonTitles-131K',
            'train': 'train.txt',
            'test': 'test.txt',
            'file_format': 'libsvm',
        }
    },
    'lf-amazon-131k': {
        'name': 'LF-Amazon-131K',
        'formats': ['bow'],
        'subsets': ['train', 'test'],
        'bow': {
            'url': 'https://drive.google.com/uc?export=download&id=1YNGEifTHu4qWBmCaLEBfjx07qRqw9DVW',  # XMLC repo url
            'dir': 'LF-Amazon-131K',
            'train': 'train.txt',
            'test': 'test.txt',
            'file_format': 'libsvm',
        }
    },
    'lf-wikiseealsotitles-320k': {
        'name': 'LF-WikiSeeAlsoTitles-320K',
        'formats': ['bow'],
        'subsets': ['train', 'test'],
        'bow': {
            'url': 'https://drive.google.com/uc?export=download&id=1edWtizAFBbUzxo9Z2wipGSEA9bfy5mdX',  # XMLC repo url
            'dir': 'LF-WikiSeeAlsoTitles-320K',
            'train': 'train.txt',
            'test': 'test.txt',
            'file_format': 'libsvm',
        }
    },
    'lf-wikiseealso-320k': {
        'name': 'LF-WikiSeeAlso-320K',
        'formats': ['bow'],
        'subsets': ['train', 'test'],
        'bow': {
            'url': 'https://drive.google.com/uc?export=download&id=1N8C_RL71ErX6X92ew9h8qRuTWJ9LywE8',  # XMLC repo url
            'dir': 'LF-WikiSeeAlso-320K',
            'train': 'train.txt',
            'test': 'test.txt',
            'file_format': 'libsvm',
        }
    },
    'lf-wikititles-500k': {
        'name': 'LF-WikiTitles-500K',
        'formats': ['bow'],
        'subsets': ['train', 'test'],
        'bow': {
            'url': 'https://drive.google.com/uc?export=download&id=1qa1HTTD509J5r4yNAH-Aq6wArljgg4mx',  # XMLC repo url
            'dir': 'LF-WikiTitles-500K',
            'train': 'train.txt',
            'test': 'test.txt',
            'file_format': 'libsvm',
        }
    },
    'lf-amazontitles-1.3m': {
        'name': 'LF-AmazonTitles-1.3M',
        'formats': ['bow'],
        'subsets': ['train', 'test'],
        'bow': {
            'url': 'https://drive.google.com/uc?export=download&id=1Davc6BIfoTIAS3mP1mUY5EGcGr2zN2pO',  # XMLC repo url
            'dir': 'LF-AmazonTitles-1.3M',
            'train': 'train.txt',
            'test': 'test.txt',
            'file_format': 'libsvm',
        }
    },
}


# Main functions for downloading and loading datasets
def load_libsvm_file(file, labels_format="list", sort_indices=False):
    """
    Load data in the libsvm format into sparse CSR matrix.
    The format is text-based. Each line contains an instance and is ended by a ``\\n`` character.

    .. code::

        <label>,<label>,... <feature>(:<value>) <feature>(:<value>) ...

    ``<label>`` and ``<feature>`` are indexes that should be positive integers.
    This method supports less-strict versions of the format.
    Labels and features do not have to be sorted in ascending order.
    The ``:<value>`` can be omitted after ``<feature>``, to assume value = 1.
    It automatically detects header used in format of datasets from
    `The Extreme Classification Repository <https://manikvarma.github.io/downloads/XC/XMLRepository.html>`_,

    :param file: Path to a file to load
    :type file: str
    :param labels_format: Format in which load the labels data (``'list'`` or ``'csr_matrix'``), defaults to `csr_matrix`
    :type labels_format: str
    :param sort_indices: If True, sort indices, otherwise keep original order, defaults to True
    :type sort_indices: bool
    :return:  Features and labels data
    :rtype: (csr_matrix, list[list[int]]) or (csr_matrix, csr_matrix)
    """
    if labels_format == 'list':
        labels, features = _load_libsvm_file_labels_list(file, sort_indices)
        return csr_matrix(features), labels
    elif labels_format == 'csr_matrix':
        labels, features = _load_libsvm_file_labels_csr_matrix(file, sort_indices)
        return csr_matrix(features), csr_matrix(labels)
    else:
        raise ValueError("Label format {} is not valid format".format(labels_format))


def load_json_lines_file(file, features_fields=['title', 'content'], labels_field='target_ind', gzip_file=None):
    """
    Load data in the JSON lines format into list of features and list of labels.
    
    :param file: Path to a JSON lines file to load
    :type file: str
    :param features_fields: list of fields of JSON line that contain features, fields will be concatenated in the specified order, defaults to ['title', 'content']
    :type features_fields: list[str], optional
    :param labels_field: field name that contains labels, defaults to ``'target_ind'``
    :type labels_field: str, optional
    :param gzip_file: If True, read file as gzip file, if None, decide based on file extension, defaults to None
    :type gzip_file: bool, optional
    :return: Raw text of documents and labels
    :rtype: (list[str], list[list[int|str]])
    """

    X = []
    Y = []
    if gzip_file == True or file[-3:] == '.gz':
        f = gzip.open(file, 'rb')
    else:
        f = open(file, 'r')
    for line in f:
        data = json.loads(line)
        if not all(f in data for f in features_fields):
            raise ValueError("Not all features fields {} are not in {}".format(features_fields, data))
        if not labels_field in data:
            raise ValueError("Labels field {} is not not in {}".format(labels_field, data))
        X.append(' '.join([data[f] for f in features_fields]))
        Y.append(data[labels_field])

    return X, Y


def download_dataset(dataset, subset='train', format='bow', root='./data', verbose=False):
    """
    Downloads the dataset from the internet and puts it in root directory.
    If dataset is already downloaded, it is not downloaded again.

    :param dataset: Name of the dataset to load, case insensitive, available datasets:

        - ``'Eurlex-4K'`` (``'bow'`` format only),
        - ``'Eurlex-4.3K'`` (``'bow'`` format only),
        - ``'AmazonCat-13K'``,
        - ``'AmazonCat-14K'``,
        - ``'Wiki10-31K'`` (alias: ``'Wiki10'``, ``'bow'`` format only),
        - ``'DeliciousLarge-200K'`` (alias: ``'DeliciousLarge'``, ``'bow'`` format only)
        - ``'WikiLSHTC-325K'`` (alias: ``'WikiLSHTC'``, ``'bow'`` format only)
        - ``'WikiSeeAlsoTitles-350K'``,
        - ``'WikiTitles-500K'``,
        - ``'WikipediaLarge-500K'`` (alias: ``'WikipediaLarge'``),
        - ``'AmazonTitles-670K'``,
        - ``'Amazon-670K'``,
        - ``'AmazonTitles-3M'``,
        - ``'Amazon-3M'``,
        - ``'LF-AmazonTitles-131K'`` (for now ``'bow'`` format only),
        - ``'LF-Amazon-131K'`` (for now ``'bow'`` format only),
        - ``'LF-WikiSeeAlsoTitles-320K'`` (for now ``'bow'`` format only),
        - ``'LF-WikiSeeAlso-320K'`` (for now ``'bow'`` format only),
        - ``'LF-WikiTitles-500K'`` (for now ``'bow'`` format only),
        - ``'LF-AmazonTitles-1.3M'`` (for now ``'bow'`` format only).

    :type dataset: str
    :param subset: Subset of dataset to download {``'train'``, ``'test'``, ``'validation'``}, defaults to ``'train'``
    :type subset: str, optional
    :param format: Format of dataset to load {``'bow'`` (bag-of-words/tf-idf weights, alias ``'tf-idf'``), ``'raw'`` (raw text)}, defaults to ``'bow'``
    :type format: str, optional
    :param root: Location of datasets directory, defaults to ``'./data'``
    :type root: str, optional
    :param verbose: If True print downloading and loading progress, defaults to False
    :type verbose: bool, optional
    """
    dataset_meta = _get_data_meta(dataset, subset=subset, format=format)
    dataset_dest = path.join(root, dataset.lower() + '_' + format + ".zip")
    data_dir = path.join(root, dataset_meta['dir'])
    file_path = dataset_meta[subset]

    if isinstance(file_path, str):
        file_path = [file_path]
    elif isinstance(file_path, dict):
        file_path = file_path.values()
    if not all(path.exists(path.join(data_dir, f)) for f in file_path):
        if 'drive.google.com' in dataset_meta['url']:
            _download_file_from_google_drive(dataset_meta['url'], dataset_dest, unzip=True, overwrite=True, delete_zip=True, verbose=verbose)


def load_dataset(dataset, subset='train', format='bow', root='./data', verbose=False):
    """
    Downloads the dataset from the internet and puts it in root directory.
    If dataset is already downloaded, it is not downloaded again.
    Then loads requested datasets into features matrix and labels.

    :param dataset: Name of the dataset to load, case insensitive, available datasets:

        - ``'Eurlex-4K'`` (``'bow'`` format only),
        - ``'Eurlex-4.3K'`` (``'bow'`` format only),
        - ``'AmazonCat-13K'``,
        - ``'AmazonCat-14K'``,
        - ``'Wiki10-31K'`` (alias: ``'Wiki10'``, ``'bow'`` format only),
        - ``'DeliciousLarge-200K'`` (alias: ``'DeliciousLarge'``, ``'bow'`` format only)
        - ``'WikiLSHTC-325K'`` (alias: ``'WikiLSHTC'``, ``'bow'`` format only)
        - ``'WikiSeeAlsoTitles-350K'``,
        - ``'WikiTitles-500K'``,
        - ``'WikipediaLarge-500K'`` (alias: ``'WikipediaLarge'``),
        - ``'AmazonTitles-670K'``,
        - ``'Amazon-670K'``,
        - ``'AmazonTitles-3M'``,
        - ``'Amazon-3M'``,
        - ``'LF-AmazonTitles-131K'`` (for now ``'bow'`` format only),
        - ``'LF-Amazon-131K'`` (for now ``'bow'`` format only),
        - ``'LF-WikiSeeAlsoTitles-320K'`` (for now ``'bow'`` format only),
        - ``'LF-WikiSeeAlso-320K'`` (for now ``'bow'`` format only),
        - ``'LF-WikiTitles-500K'`` (for now ``'bow'`` format only),
        - ``'LF-AmazonTitles-1.3M'`` (for now ``'bow'`` format only).

    :type dataset: str
    :param subset: Subset of dataset to load into features matrix and labels {``'train'``, ``'test'``, ``'validation'``}, defaults to ``'train'``
    :type subset: str, optional
    :param format: Format of dataset to load {``'bow'`` (bag-of-words/tf-idf weights, alias ``'tf-idf'``), ``'raw'`` (raw text)}, defaults to ``'bow'``
    :type format: str, optional
    :param root: Location of datasets directory, defaults to ``'./data'``
    :type root: str, optional
    :param verbose: If True print downloading and loading progress, defaults to False
    :type verbose: bool, optional
    :return: Tuple of features matrix and labels.
    :rtype: (csr_matrix, list[list[int]]) or (list[str], list[list[str]])
    """
    download_dataset(dataset, subset=subset, format=format, root=root, verbose=verbose)
    dataset_meta = _get_data_meta(dataset, subset=subset, format=format)
    file_format = dataset_meta['file_format']
    data_dir = path.join(root, dataset_meta['dir'])
    file_path = dataset_meta[subset]

    if file_format == 'libsvm':
        return load_libsvm_file(path.join(data_dir, file_path))
    elif file_format == 'XY_sparse':
        X, _ = load_libsvm_file(path.join(data_dir, file_path['X']))
        Y, _ = load_libsvm_file(path.join(data_dir, file_path['Y']))
        return X, Y
    elif file_format == 'jsonlines':
        return load_json_lines_file(path.join(data_dir, file_path), features_fields=dataset_meta['features_fields'], labels_field=dataset_meta['labels_field'])
    else:
        raise ValueError("File format {} is not supported".format(file_format))


def to_csr_matrix(X, shape=None, sort_indices=False, dtype=np.float32):
    """
    Converts sparse matrix-like data, like list of list of tuples (idx, value), to Scipy csr_matrix.

    :param X: Matrix-like object to convert to csr_matrix: ndarray or list of lists of ints or tuples of ints and floats (idx, value).
    :type X: ndarray, list[list[int|str]], list[list[tuple[int, float]]
    :param shape: Shape of the matrix, if None, shape will be deduce from X, defaults to None
    :type shape: tuple, optional
    :param sort_indices: Sort rows' data by indices (idx), defaults to False
    :type sort_indices: bool, optional
    :param dtype: Data type of the matrix, defaults to np.float32
    :type dtype: type, optional
    :return: X as csr_matrix.
    :rtype: csr_matrix
    """
    if isinstance(X, list) and isinstance(X[0], (list, tuple, set)):
        first_element = _get_first_element_of_list_of_lists(X)

        if dtype is None:
            dtype = _get_dtype(first_element)

        size = 0
        for x in X:
            size += len(x)

        indptr = np.zeros(len(X) + 1, dtype=np.int32)
        indices = np.zeros(size, dtype=np.int32)
        data = np.ones(size, dtype=dtype)
        cells = 0

        if isinstance(first_element, tuple):
            for row, x in enumerate(X):
                indptr[row] = cells
                x = sorted(x) if sort_indices else x
                for x_i in x:
                    indices[cells] = x_i[0]
                    data[cells] = x_i[1]
                    cells += 1
            indptr[len(X)] = cells

        else:
            for row, x in enumerate(X):
                indptr[row] = cells
                indices[cells:cells + len(x)] = sorted(x) if sort_indices else x
                cells += len(x)
            indptr[len(X)] = cells

        array = csr_matrix((data, indices, indptr), shape=shape)
    else: # Try to convert via constructor
        array = csr_matrix(X, dtype=dtype, shape=shape)
    # raise TypeError('Cannot convert X to csr_matrix')

    # Check type
    if array.dtype != dtype:
        print("Conversion", array.dtype, dtype)
        array = array.astype(dtype)

    return array

def to_np_matrix(X, shape=None, dtype=np.float32):
    """
    Converts sparse matrix-like data, like list of list of tuples (idx, value), to Numpy matrix (2D array).

    :param X: Matrix-like object to convert to csr_matrix: ndarray or list of lists of ints or tuples of ints and floats (idx, value).
    :type X: ndarray, list[list[int|str]], list[list[tuple[int, float]]
    :param shape: Shape of the matrix, if None, shape will be deduce from X, defaults to None
    :type shape: tuple, optional
    :param dtype: Data type of the matrix, if None, type will be deduced from data, defaults to np.float32
    :type dtype: type, optional
    :return: X as Numpy 2D array.
    :rtype: ndarray
    """
    if isinstance(X, list) and isinstance(X[0], (list, tuple, set)):
        first_element = _get_first_element_of_list_of_lists(X)

        if dtype is None:
            dtype = _get_dtype(first_element)

        if shape is None:
            m = max([max(x) for x in X if len(x)]) + 1
            n = len(X)
            shape = (n, m)
        array = np.zeros(shape, dtype=dtype)

        if isinstance(first_element, (list, tuple)):
            for row, x in enumerate(X):
                for x_i in x:
                    array[row, x_i[0]] = x_i[1]
        else:
            for row, x in enumerate(X):
                array[row, x] = 1

    elif isinstance(X, csr_matrix):
        array = X.toarray()
    else: # Try to convert via constructor
        array = np.array(X, dtype=dtype)

    # Check type
    if array.dtype != dtype:
        print("Conversion", array.dtype, dtype)
        array = array.astype(dtype)

    return array


# Helpers
def _get_dtype(first_element):
    if isinstance(first_element, (list, tuple)):
        return type(first_element[1])
    else:
        return type(first_element)


def _get_first_element_of_list_of_lists(X):
    first_element = None
    for x in X:
        if len(x):
            first_element = x[0]
    if first_element is None:
        raise ValueError('X is does not contain any element')


def _get_data_meta(dataset, subset='train', format='bow'):
    # Handel aliases
    aliases = {
        'wiki10': 'wiki10-31k',
        'deliciouslarge': 'deliciouslarge-200k',
        'wikilshtc': 'wikilshtc-325k',
        'wikipedialarge': 'wikipedialarge-500k'
    }
    _dataset = dataset.lower()
    if _dataset in aliases:
        _dataset = aliases[_dataset]

    if _dataset not in DATASETS:
        raise ValueError("Dataset {} is not available".format(dataset))

    # Handel format aliases and versions
    _format = format
    if _format == 'tf-idf':
        _format = 'bow'
    if _format.startswith("bow") and _format not in DATASETS[_dataset]:
        _format = 'bow'

    if _format not in DATASETS[_dataset]['formats']:
        raise ValueError("Format {} is not available for dataset {}".format(format, dataset))

    if subset is not None and subset not in DATASETS[_dataset]['subsets']:
        raise ValueError("Subset {} is not available for dataset {}".format(format, dataset))

    return DATASETS[_dataset][_format]


def _download_file_from_google_drive(url, dest_path, overwrite=False, unzip=False, delete_zip=False, verbose=False):
    """
    Downloads a shared file from google drive into a given folder and optionally unzips it.

    :param url: File url to download
    :type url: str
    :param dest_path: The destination where to save the downloaded file
    :type dest_path: str
    :param overwrite: If True force redownload and overwrite, defaults to False
    :type overwrite: bool
    :param unzip: If True unzip a file, optional, defaults to False
    :type unzip: bool
    :param delete_zip: If True and unzips is True delete archive file after unziping, defaults to False
    :type delete_zip: bool
    :param verbose: If True print downloading progress, defaults to False
    :type verbose: bool
    :return: None
    """

    destination_directory = path.dirname(dest_path)
    if not path.exists(destination_directory):
        makedirs(destination_directory)

    if not path.exists(dest_path) or overwrite:
        if verbose:
            print('Downloading {} into {} ... '.format(url, dest_path))

        gdown.download(url=url, output=dest_path, quiet=not verbose)

        if unzip:
            try:
                if verbose:
                    print('Unzipping ...')

                with zipfile.ZipFile(dest_path, 'r') as z:
                    z.extractall(destination_directory)
                if delete_zip:
                    remove(dest_path)

            except zipfile.BadZipfile:
                warnings.warn('Ignoring `unzip` since "{}" does not look like a valid zip file'.format(dest_path))

        if verbose:
            print('Done.')




import shutil
from napkinxc.datasets import download_dataset

from conf import *


def pytest_configure(config):
    print("Downloading/checking test data...")
    download_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)
    download_dataset(TEST_DATASET, "test", root=TEST_DATA_PATH)


def pytest_unconfigure(config):
    if REMOVE_TEST_DATA:
        print("Removing test data...")
        shutil.rmtree(TEST_DATA_PATH, ignore_errors=True)



from time import time
import numpy as np
import shutil

# pip install git+https://github.com/kunaldahiya/pyxclib.git
from xclib.evaluation.xc_metrics import *

from napkinxc.datasets import load_dataset, to_csr_matrix
from napkinxc.models import PLT
from napkinxc.measures import *


from conf import *
MODEL_PATH = get_model_path(__file__)


def test_compare_napkinxc_with_xclib():
    k = 5
    
    # Train model and predict
    X_train, Y_train = load_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)
    X_test, Y_test = load_dataset(TEST_DATASET, "test", root=TEST_DATA_PATH)
    plt = PLT(MODEL_PATH)
    plt.fit(X_train, Y_train)
    Y_pred = plt.predict_proba(X_test, top_k=k)
    shutil.rmtree(MODEL_PATH, ignore_errors=True)

    # Prepare dataset
    csr_Y_train = to_csr_matrix(Y_train)
    csr_Y_test = to_csr_matrix(Y_test)
    csr_Y_pred = to_csr_matrix(Y_pred, shape=csr_Y_test.shape)

    # Calculate propensities
    nxc_inv_ps = inverse_propensity(Y_train, A=0.55, B=1.5)
    csr_nxc_inv_ps = inverse_propensity(csr_Y_train, A=0.55, B=1.5)
    xcl_inv_ps = compute_inv_propesity(csr_Y_train, A=0.55, B=1.5)
    assert np.allclose(nxc_inv_ps, csr_nxc_inv_ps)
    assert np.allclose(nxc_inv_ps, xcl_inv_ps)

    # Compare results
    measures = {
        "P@k": {"nxc": precision_at_k, "xclib": precision, "inv_ps": False},
        "R@k": {"nxc": recall_at_k, "xclib": recall, "inv_ps": False},
        "nDCG@k": {"nxc": ndcg_at_k, "xclib": ndcg, "inv_ps": False},
        "PSP@k": {"nxc": psprecision_at_k, "xclib": psprecision, "inv_ps": True},
        "PSR@k": {"nxc": psrecall_at_k, "xclib": psrecall, "inv_ps": True},
        "PSnDCG@k": {"nxc": psndcg_at_k, "xclib": psndcg, "inv_ps": True}
    }

    print("\n")
    for m, v in measures.items():
        print("\n{} time comparison:".format(m))
        t_start = time()
        xclib_r = v["xclib"](csr_Y_pred, csr_Y_test, xcl_inv_ps, k=k) if v["inv_ps"] else v["xclib"](csr_Y_pred, csr_Y_test, k=k)
        print("\txclib.evaluation.xc_metrics.{} with csr_matrices: {}s".format(v["xclib"].__name__, time() - t_start))

        t_start = time()
        nxc_r = v["nxc"](Y_test, Y_pred, xcl_inv_ps, k=k) if v["inv_ps"] else v["nxc"](Y_test, Y_pred, k=k)
        print("\tnapkinXC.measures.{} with lists: {}s".format(v["nxc"].__name__, time() - t_start))

        t_start = time()
        csr_nxc_r = v["nxc"](csr_Y_test, csr_Y_pred, csr_nxc_inv_ps, k=k) if v["inv_ps"] else v["nxc"](csr_Y_test, csr_Y_pred, k=k)
        print("\tnapkinXC.measures.{} with csr_matrices: {}s".format(v["nxc"].__name__, time() - t_start))

        assert np.allclose(nxc_r, csr_nxc_r)
        assert np.allclose(nxc_r, xclib_r)



#!/usr/bin/env python3

# This demo shows how to train, store and later load the napkinXC model.

from napkinxc.datasets import load_dataset
from napkinxc.models import PLT
from napkinxc.measures import precision_at_k

# The beginning is the same as in the basic.py example.

# Use load_dataset function to load one of the benchmark datasets
# from XML Repository (http://manikvarma.org/downloads/XC/XMLRepository.html).
X_train, Y_train = load_dataset("eurlex-4k", "train")
X_test, Y_test = load_dataset("eurlex-4k", "test")

# Create PLT model with "eurlex-model" directory,
# it will be created and used during model training for storing weights.
# napkinXC stores already trained parts of the models to save RAM.
plt = PLT("eurlex-model")

# Fit the model on the training dataset.
# The model weights and additional data will be stored in "eurlex-model" directory.
plt.fit(X_train, Y_train)

# Predict.
Y_pred = plt.predict(X_test, top_k=1)
print("Precision at 1:", precision_at_k(Y_test, Y_pred, k=1))

# Delete plt object.
del plt

# To load the model, create a new PLT object with the same directory as the previous one.
new_plt = PLT("eurlex-model")

# Predict using a new model object.
Y_pred = new_plt.predict(X_test, top_k=1)
print("Precision at 1 after loading:", precision_at_k(Y_test, Y_pred, k=1))



import os

TEST_DATA_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "test_data")
TEST_DATASET = "yeast"  # old: "eurlex-4k"
REMOVE_TEST_DATA = False
TEST_SEED = 1993
SCORE_RANGE = [0.61, 0.77]  # old for eurlex-4k: [0.72, 0.82]


def get_model_path(test_file):
    return os.path.join(os.path.dirname(os.path.abspath(test_file)), f"{os.path.basename(test_file)}_model")




from time import time
import os
from napkinxc.datasets import *

from conf import *


def test_load_dataset():
    for d in DATASETS.values():
        for f in d['formats']:
            for s in d['subsets']:
                download_dataset(d['name'], subset=s, format=f, root=TEST_DATA_PATH)
                t_start = time()
                X, Y = load_dataset(d['name'], subset=s, format=f, root=TEST_DATA_PATH)
                len_X = len(X) if isinstance(X, list) else X.shape[0]
                len_Y = len(Y) if isinstance(Y, list) else Y.shape[0]
                assert len_X == len_Y
                print("\tload_dataset({}, subset={}, format={}) rows: {}, time: {}s\n".format(d['name'], s, f, len_X, time() - t_start))



from time import time
import os
from sklearn.datasets import load_svmlight_file
from napkinxc.datasets import download_dataset, load_libsvm_file
import numpy as np

from conf import *
MODEL_PATH = get_model_path(__file__)


def test_load_libsvm():
    datasets = {
        "eurlex-4k": {"file": os.path.join(TEST_DATA_PATH, "Eurlex/eurlex_test.txt"), "sklearn_args": {"multilabel": True, "zero_based": True, "n_features": 5000, "offset": 1}},
        "amazonCat-13k": {"file": os.path.join(TEST_DATA_PATH, "AmazonCat/amazonCat_test.txt"), "sklearn_args": {"multilabel": True, "zero_based": True, "n_features": 203882, "offset": 1}},
        "amazonCat-14k": {"file": os.path.join(TEST_DATA_PATH, "AmazonCat-14K/amazonCat-14K_test.txt"), "sklearn_args": {"multilabel": True, "zero_based": True, "n_features": 597540, "offset": 1}},
        "wiki10-31k": {"file": os.path.join(TEST_DATA_PATH, "Wiki10/wiki10_test.txt"), "sklearn_args": {"multilabel": True, "zero_based": True, "n_features": 101938, "offset": 1}}
    }

    for d, v in datasets.items():
        download_dataset(d, subset='test', format='bow', root=TEST_DATA_PATH)
        print("\n{} time comparison:".format(d))

        t_start = time()
        sk_X, sk_Y = load_svmlight_file(v["file"], **v["sklearn_args"])
        print("\tsklearn.datasets.load_svmlight_file time: {}s".format(time() - t_start))

        t_start = time()
        nxc_X1, nxc_Y_list = load_libsvm_file(v["file"], labels_format="list")
        print("\tnapkinXC.datasets.load_libsvm_file time: {}s".format(time() - t_start))

        t_start = time()
        nxc_X2, nxc_Y_csrm = load_libsvm_file(v["file"], labels_format="csr_matrix")
        print("\tnapkinXC.datasets.load_libsvm_file time: {}s".format(time() - t_start))

        assert np.array_equal(nxc_X1.indptr, nxc_X2.indptr)
        assert np.array_equal(nxc_X1.indices, nxc_X2.indices)
        assert np.array_equal(nxc_X1.data, nxc_X2.data)

        assert np.array_equal(nxc_X1.indptr, sk_X.indptr)
        assert np.array_equal(nxc_X1.indices, sk_X.indices)
        assert np.allclose(nxc_X1.data, sk_X.data)
        assert nxc_X1.shape[0] == nxc_Y_csrm.shape[0]

        assert len(nxc_Y_list) == len(sk_Y)
        for nxc_y, sk_y in zip(nxc_Y_list, sk_Y):
            assert len(nxc_y) == len(sk_y)
            assert all(y1 == y2 for y1, y2 in zip(nxc_y, sk_y))



# Copyright (c) 2020-2021 by Marek Wydmuch
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
from numpy import ndarray
from scipy.sparse import csr_matrix
from ._napkinxc import CPPModel, InputDataType


class Model():
    """
    Main model class that wraps CPPModel
    """

    def __init__(self, **params):
        self._model = CPPModel() # In the feature more CPP classes may be available
        self._params = {}
        self.set_params(**params)

    def fit(self, X, Y):
        """
        Fit the model to the given training data.

        :param X: Training data points as a matrix or list of lists of int or tuples of int and float (feature id, value).
        :type X: csr_matrix, ndarray, list[list[int]|tuple[int]], list[list[tuple[int, float]]
        :param Y: Target labels as a matrix or lists or tuples of ints (multi-label data) or list of ints (multi-class data).
        :type Y: csr_matrix|ndarray|list[list[int]|tuple[int]], list[list[tuple[int, float]], list[int]
        """
        self._model.fit(X, Y, Model._check_data_type(X), Model._check_data_type(Y))

    def fit_on_file(self, path):
        """
        Fit the model to the training data in the given file in multi-label svmlight/libsvm format.

        :param path: Path to the file.
        :type path: str
        """
        self._model.fitOnFile(path)

    def load(self):
        """
        Load the model to RAM.
        """
        self._model.load()

    def unload(self):
        """
        Unload the model from RAM.
        """
        self._model.unload()

    def predict(self, X, top_k=0, threshold=0, labels_weights=None):
        """
        Predict labels for data points in X.

        :param X: Data points as a matrix or list of lists of int or tuples of int and float (feature id, value).
        :type X: csr_matrix, ndarray, list[list[int]|tuple[int]], list[list[tuple[int, float]]
        :param top_k: Predict top-k labels, if 0, the option is ignored, defaults to 0
        :type top_k: int
        :param threshold: Predict labels with probability above the threshold in case of single value
            or above the specific threshold for each label in case of list or array of values,
            if 0, the option is ignored, defaults to 0
        :type threshold: float, list[float], ndarray, optional
        :param labels_weights: Predict labels according to their weights multiplied by probability
            if None, the option is ignored, defaults to None
        :type labels_weights: list[float], ndarray, optional
        :return: List of lists with predicted labels.
        :rtype: list[list[int]]
        """
        threshold = self._prepare_pred(top_k, threshold, labels_weights)
        return self._model.predict(X, Model._check_data_type(X), top_k, threshold)

    def predict_proba(self, X, top_k=0, threshold=0, labels_weights=None):
        """
        Predict labels with probability estimates for data points in X.

        :param X: Data points as a matrix or list of lists of int or tuples of int and float (feature id, value).
        :type X: csr_matrix, ndarray, list[list[int]|tuple[int]], list[list[tuple[int, float]]
        :param top_k: Predict top-k labels, if 0, the option is ignored, defaults to 0
        :type top_k: int
        :param threshold: Predict labels with probability above the threshold in case of single value
            or above the specific threshold for each label in case of list or array of values,
            if 0, the option is ignored, defaults to 0
        :type threshold: float, list[float], ndarray, optional
        :param labels_weights: Predict labels according to their weights multiplied by probability
            if None, the option is ignored, defaults to None
        :type labels_weights: list[float], ndarray, optional
        :return: List of list of tuples (label id, probability) with predicted labels
        :rtype: list[list[tuple[int, float]]
        """
        threshold = self._prepare_pred(top_k, threshold, labels_weights)
        return self._model.predict_proba(X, Model._check_data_type(X), top_k, threshold)

    def predict_for_file(self, path, top_k=0, threshold=0, labels_weights=None):
        """
        Predict labels for data points in the given file in multi-label svmlight/libsvm format.

        :param path: Path to the file
        :type path: str
        :param top_k: Predict top-k labels, if 0, the option is ignored, defaults to 0
        :type top_k: int
        :param threshold: Predict labels with probability above the threshold in case of single value
            or above the specific threshold for each label in case of list or array of values,
            if 0, the option is ignored, defaults to 0
        :type threshold: float, list[float], ndarray, optional
        :param labels_weights: Predict labels according to their weights multiplied by probability
            if None, the option is ignored, defaults to None
        :type labels_weights: list[float], ndarray, optional
        :return: List of lists with predicted labels.
        :rtype: list[list[int]]
        """
        threshold = self._prepare_pred(top_k, threshold, labels_weights)
        return self._model.predict_for_file(path, top_k, threshold)

    def predict_proba_for_file(self, path, top_k=0, threshold=0, labels_weights=None):
        """
        Predict labels with probability estimates for data points in the given file in multi-label svmlight/libsvm format.

        :param path: Path to the file.
        :type path: str
        :param top_k: Predict top-k labels, if 0, the option is ignored, defaults to 0
        :type top_k: int
        :param threshold: Predict labels with probability above the threshold in case of single value
            or above the specific threshold for each label in case of list or array of values,
            if 0, the option is ignored, defaults to 0
        :type threshold: float, list[float], ndarray, optional
        :param labels_weights: Predict labels according to their weights multiplied by probability
            if None, the option is ignored, defaults to None
        :type labels_weights: list[float], ndarray, optional
        :return: List of list of tuples (label id, probability) with predicted labels
        :rtype: list[list[tuple[int, float]]
        """
        threshold = self._prepare_pred(top_k, threshold, labels_weights)
        return self._model.predict_proba_for_file(path, top_k, threshold)

    def ofo(self, X, Y, type='micro', a=10, b=20, epochs=1):
        """
        Perform Online F-measure Optimization procedure on the given data to find optimal thresholds.

        :param X: Data points as a matrix or list of lists of int or tuples of int and float (feature id, value).
        :type X: csr_matrix, ndarray, list[list[int]|tuple[int]], list[list[tuple[int, float]]
        :param Y: Target labels as a matrix or lists or tuples of ints (multi-label data) or list of ints (multi-class data).
        :type Y: csr_matrix, ndarray, list[list[int]|tuple[int]], list[list[tuple[int, float]], list[int]
        :param type: Type of OFO procedure {``'micro'``, ``'macro'``}, default to ``'micro'``
        :type type: str
        :param a: Parameter of OFO procedure, defaults to 10
        :type a: int
        :param b: Parameter of OFO procedure, defaults to 20
        :type b: int
        :param epochs: Number of OFO epochs, defaults to 1
        :type epochs: int, optional
        :return: Single threshold in case of ``type='micro'`` and list of thresholds in case of ``type='macro'``
        :rtype: float, list[float]
        """
        self.set_params(ofo_type=type, ofo_a=a, ofo_b=b, epochs=epochs)
        thr = self._model.ofo(X, Y, Model._check_data_type(X), Model._check_data_type(Y))
        if type == 'micro':
            thr = thr[0]
        return thr

    def get_params(self, deep=False): # deep argument for Scikit-learn compatibility
        """
        Get parameters of this model.

        :param deep: Ignored, added for Scikit-learn compatibility, defaults to False
        :return: Mapping of string to any
        :rtype: dict
        """
        return self._params

    def set_params(self, **params):
        """
        Set parameters for this model. Should be used only if you know what are you doing.

        :param: \*\*params: Parameter names with their new values.
        :return: self
        :rtype: Model
        """
        if 'model' in self._params and 'model' in params:
            params.pop('model')  # do not allow for changing model param

        if 'verbose' in params and params['verbose'] == True:
            params['verbose'] = 2  # override verbose

        self._params.update(params)
        params_list = []
        for k, v in params.items():
            if v is None:
                continue
            elif not isinstance(v, (float, int, str, bool)):
                raise TypeError("Model argument \"{}\" should be float, int, str, bool or None".format(k))

            arg_k = ("--" if len(k) > 1 else "-") + Model._to_camelcase(k)
            arg_v = v
            if type(arg_v) == bool:
                arg_v = int(arg_v)
            arg_v = str(arg_v)

            params_list.extend([arg_k, arg_v])
        self._model.set_args(params_list)

        return self

    # def get_weights(self):
    #     """
    #     Returns weights of binary classifiers used in the model.
    #
    #     :return: Sparse matrix with weights.
    #     :rtype: csr_matrix
    #     """
    #     weights = self._model.get_weights()
    #     return csr_matrix(weights)
    #
    # def set_weights(self, W):
    #     """
    #     Set weights of binary classifiers used in the model.
    #
    #     :param W: Matrix with weights.
    #     :type W: csr_matrix|ndarray
    #     """
    #     self._model.set_weights(W, Model._check_data_type(W))

    @staticmethod
    def _get_init_params(locals):
        kwargs = locals['kwargs']
        for k in ['self', '__class__', 'kwargs']:
            locals.pop(k)
        locals.update(kwargs)
        return locals

    @staticmethod
    def _to_camelcase(string):
        return re.sub(r'(?!^)_([a-zA-Z])', lambda m: m.group(1).upper(), string)

    @staticmethod
    def _check_data_type(data):
        if isinstance(data, list):
            return InputDataType.list
        elif isinstance(data, ndarray):
            return InputDataType.ndarray
        elif isinstance(data, csr_matrix):
            return InputDataType.csr_matrix
        else:
            return -1

    def _prepare_pred(self, top_k, threshold, labels_weights):
        if top_k == 0 and threshold == 0:
            print("Warning: both top_k and threshold arguments set to 0, this will predict all labels")

        if not isinstance(top_k, int):
            raise TypeError("Unsupported top_k type, should be int")

        if isinstance(threshold, (list, ndarray)):
            self._model.set_thresholds(threshold)
            threshold = 0
        elif not isinstance(threshold, (float, int)):
            raise TypeError("Unsupported threshold type, should be float, or list of floats, or Numpy vector (1d array)")

        if isinstance(labels_weights, (list, ndarray)):
            self._model.set_labels_weights(labels_weights)
        elif labels_weights is not None:
            raise TypeError("Unsupported labels_weights type, should be list of floats, or Numpy vector (1d array)")

        return threshold


class LabelTreeModel(Model):
    """
    Base class for label tree-based models
    """

    def __init__(self, **params):
        tree_structure = params["tree_structure"]
        if isinstance(tree_structure, list):
            params.pop("tree_structure")
        elif tree_structure is not None and not isinstance(tree_structure, str):
            raise TypeError("Model argument \"tree_structure\" should be list, str (filepath) or None")

        super(LabelTreeModel, self).__init__(**params)
        if isinstance(tree_structure, list):
            self.set_tree_structure(tree_structure)

    def build_tree(self, X, Y):
        """
        Build the tree for the given data (without training node classifiers)

        :param X: Data points as a matrix or list of lists of int or tuples of int and float (feature id, value).
        :type X: ndarray, csr_matrix, list[list[int]], list[list[tuple[int, float]]
        :param Y: Target labels as list of ints (multi-class data) or lists or tuples of ints (multi-label data).
        :type Y: list[int], list[list|tuple[int]]
        """
        self._model.build_tree(X, Y, Model._check_data_type(X), Model._check_data_type(Y))

    def get_tree_structure(self):
        """
        Return internal label tree structure

        :return: Tree structure, represented as a list of tuples representing nodes,
            where the first value is an index of a parent node, if equal to -1, then the node is a root node,
            the second value is an index of the node,
            and the third, a label assigned to the node, if equal to -1, then no label is assigned to the node.
        :rtype: list[tuple[int, int, int]]
        """
        return self._model.get_tree_structure()

    def set_tree_structure(self, tree_structure):
        """
        Set internal label tree structure

        :param tree_structure: Tree structure in format of a list of tuples representing nodes,
            where the first value is an index of a parent node, if equal to -1, then the node is a root node,
            the second value is an index of the node,
            and the third, a label assigned to the node, if equal to -1, then no label is assigned to the node.
        :type tree_structure: list[tuple[int, int, int]]
        """
        self._model.set_tree_structure(tree_structure)

    def get_nodes_to_update(self, Y):
        """
        Based on the current tree, get list of updates for each set of labels in Y.

        :param Y: Target labels as list of ints (multi-class data) or lists or tuples of ints (multi-label data).
        :type Y: list[int], list[list|tuple[int]]
        :return: List of lists of nodes and their updates (0 - negative  or 1 - positive) for each set of labels in Y.
        :rtype: list[list[tuple[int, float]]]
        """
        return self._model.get_nodes_to_update(Y, self._check_data_type(Y))

    def get_nodes_updates(self, Y):
        """
        Based on the current tree, get list of updates for each node for dataset Y.

        :param Y: Target labels as list of ints (multi-class data) or lists or tuples of ints (multi-label data).
        :type Y: list[int], list[list[int]|tuple[int]]
        :return: List of lists of examples and their updates (0 - negative  or 1 - positive) for each node in the current tree.
        :rtype: list[list[tuple[int, float]]]
        """
        return self._model.get_nodes_updates(Y, self._check_data_type(Y))

    @staticmethod
    def remap_tree_structure(tree_structure):
        """
        Remaps tree structure to list of tuples of ints.

        :param tree_structure: Tree structure in format of a list of tuples representing nodes,
            where the first value is name/index a parent node, if equal to None or -1, then the node is a root node,
            the second value is name/index of the node,
            and the third, a label assigned to the node, if equal to None or -1, then no label is assigned to the node.
        :type tree_structure: list[tuple[any, any, any]]
        :return: Tree structure, represented as a list of tuples representing nodes,
            where the first value is an index of a parent node, if equal to -1, then the node is a root node,
            the second value is an index of the node,
            and the third, a label assigned to the node, if equal to -1, then no label is assigned to the node.
        :rtype: list[tuple[int, int, int]]
        """
        def _get_remapped_index(map, index):
            remapped_index = map.get(index, None)
            if remapped_index is None:
                remapped_index = len(map)
                map[index] = remapped_index
            return remapped_index

        remapped_tree_structure = []
        labels_map = {}
        nodes_map = {}
        for (p, n, l) in tree_structure:
            rp = _get_remapped_index(nodes_map, p)
            np = _get_remapped_index(nodes_map, p)
            lp = _get_remapped_index(labels_map, l)
        return remapped_tree_structure


class PLT(LabelTreeModel):
    """
    Probabilistic Labels Trees (PLTs) (multi-label) classifier with linear node estimators, using CPP core.
    """

    def __init__(self,
                 output,

                 # Tree params
                 tree_type='hierarchicalKmeans',
                 arity=2,
                 max_leaves=100,
                 kmeans_eps=0.0001,
                 kmeans_balanced=True,
                 flatten_tree=0,
                 tree_structure=None,

                 # Features params
                 hash=None,
                 features_threshold=0,
                 norm=True,
                 bias=1.0,

                 # Base (node) classifiers params
                 optimizer='liblinear',
                 loss='log',
                 weights_threshold=0.1,
                 liblinear_c=10,
                 liblinear_eps=0.1,
                 liblinear_solver=None,
                 liblinear_max_iter=100,
                 eta=1.0,
                 epochs=1,
                 adagrad_eps=0.001,

                 # Prediction
                 tree_search_type='exact',
                 beam_search_width=10,
                 load_as='map',

                 # Other
                 ensemble=1,
                 seed=None,
                 threads=0,
                 verbose=0,
                 **kwargs):
        """
        Construct a Probabilistic Labels Trees model.

        :param output: Directory where the model will be stored
        :type output: str
        :param tree_type: Tree type to construct. Available tree types:

            - ``'hierarchicalKmeans'``
            - ``'balancedInOrder'``
            - ``'balancedRandom'``
            - ``'completeKaryInOrder'``
            - ``'completeKaryRandom'``
            - ``'huffman'``

            Defaults to ``'hierarchicalKmeans'``
        :type tree_type: str, optional
        :param arity: Arity of tree nodes, k for k-means clustering used in hierarchical k-means tree building procedure, defaults to 2
        :type arity: int, optional
        :param max_leaves: Maximum degree of pre-leaf nodes, defaults to 100
        :type max_leaves: int, optional
        :param kmeans_eps: Tolerance of termination criterion of the k-means clustering used in hierarchical k-means tree building procedure, defaults to 0.0001
        :type kmeans_eps: float, optional
        :param kmeans_balanced: Use balanced k-means clustering, defaults to True
        :type kmeans_balanced: bool, optional
        :param tree_search_type: Tree search algorithm used for prediction {``'exact'``, ``'beam'``}, defaults to ``'exact'``
        :type tree_search_type: str, optional
        :param beam_search_width: Width of the tree beam search, makes effect only if ``tree_search_type='beam'``, defaults to 10
        :type beam_search_width: int, optional
        :param hash: Hash features to a space of given size, value of this argument is saved with model weights, if None or 0 disable hashing, defaults to None
        :type hash: int, optional
        :param features_threshold: Prune features below given threshold, value of this argument is saved with model weights, defaults to 0
        :type features_threshold: float, optional
        :param norm: Unit norm feature vector, value of this argument is saved with model weights, defaults to True
        :type norm: bool, optional
        :param bias: Value of the bias features, value of this argument is saved with model weights, defaults to 1.0
        :type bias: float, optional
        :param optimizer: Optimizer used for training node classifiers {``'liblinear'``, ``'sgd'``, ``'adagrad'``}, defaults to ``'liblinear'``
        :type optimizer: str, optional
        :param loss: Loss optimized while training node classifiers {``'log'`` (alias ``'logistic'``), ``'l2'`` (alias ``'squaredHinge'``)}, defaults to ``'log'``
        :type loss: str, optional
        :param weights_threshold: Threshold value for pruning weights, defaults to 0.1
        :type weights_threshold: float, optional
        :param liblinear_c: LIBLINEAR cost co-efficient, inverse regularization strength, smaller values specify stronger regularization, makes effect only if ``optimizer='liblinear'``, defaults to 10.0
        :type liblinear_c: float, optional
        :param liblinear_eps: LIBLINEAR tolerance of termination criterion, makes effect only if ``optimizer='liblinear'``, defaults to 0.1
        :type liblinear_eps: float, optional
        :param liblinear_solver: Override LIBLINEAR solver set by loss parameter (default for ``loss='log'``: ``'L2R_LR_DUAL'``, for ``loss='l2'``: ``'L2R_L2LOSS_SVC_DUAL'``), makes effect only if ``optimizer='liblinear'``.
            Available solvers:

            - ``'L2R_LR_DUAL'``
            - ``'L2R_LR'``
            - ``'L1R_LR'``
            - ``'L2R_L2LOSS_SVC_DUAL'``
            - ``'L2R_L2LOSS_SVC'``
            - ``'L2R_L1LOSS_SVC_DUAL'``
            - ``'L1R_L2LOSS_SVC'``

            ``L2R_LR_DUAL`` and ``L2R_L2LOSS_SVC_DUAL`` usually work the best in XC setting, defaults to None
        :type liblinear_solver: str, optional
        :param liblinear_max_iter: Limits number of iteration by LIBLINEAR, makes effect only if ``optimizer='liblinear'``, defaults to 100
        :type liblinear_max_iter: int, optional
        :param eta: Step size (learning rate) for online optimizers, defaults to 1.0
        :type eta: float, optional
        :param epochs: Number of training epochs for online optimizers, defaults to 1
        :type epochs: int, optional
        :param adagrad_eps: Defines starting step size for AdaGrad, defaults to 0.001
        :type adagrad_eps: float, optional
        :param ensemble: Number of trees in the ensemble, defaults to 1
        :type ensemble: int, optional
        :param seed: Seed, If None use current system time, defaults to None
        :type seed: int, optional
        :param threads: Number of threads used for training and prediction, if 0 use number of available CPUs, if -1 use number of available CPUs - 1, defaults to 0
        :type threads: int, optional
        :param verbose: If True print progress, defaults to False
        :type verbose: bool, optional
        """
        all_params = Model._get_init_params(locals())
        all_params.update({"model": "plt"})
        super(PLT, self).__init__(**all_params)


class HSM(LabelTreeModel):
    """
    Hierarchical Softmax (multi-class) classifier with linear node estimators, using CPP core.
    """

    def __init__(self,
                 output,

                 # Tree params
                 tree_type='hierarchicalKmeans',
                 arity=2,
                 max_leaves=100,
                 kmeans_eps=0.0001,
                 kmeans_balanced=True,
                 flatten_tree=0,
                 tree_structure=None,

                 # Features params
                 hash=None,
                 features_threshold=0,
                 norm=True,
                 bias=1.0,
                 pick_one_label_weighting=False,

                 # Base (node) classifiers params
                 optimizer='liblinear',
                 loss='log',
                 weights_threshold=0.1,
                 liblinear_c=10,
                 liblinear_eps=0.1,
                 liblinear_solver=None,
                 liblinear_max_iter=100,
                 eta=1.0,
                 epochs=1,
                 adagrad_eps=0.001,

                 # Prediction
                 load_as='map',

                 # Other
                 ensemble=1,
                 seed=None,
                 threads=0,
                 verbose=0,
                 **kwargs):
        """
        Construct a Hierarchical Softmax model.

        :param output: Directory where the model will be stored
        :type output: str
        :param tree_type: Tree type to construct. Available tree types:

            - ``'hierarchicalKmeans'``
            - ``'balancedInOrder'``
            - ``'balancedRandom'``
            - ``'completeKaryInOrder'``
            - ``'completeKaryRandom'``
            - ``'huffman'``

            Defaults to ``'hierarchicalKmeans'``
        :type tree_type: str, optional
        :param arity: Arity of tree nodes, k for k-means clustering used in hierarchical k-means tree building procedure, defaults to 2
        :type arity: int, optional
        :param max_leaves: Maximum degree of pre-leaf nodes, defaults to 100
        :type max_leaves: int, optional
        :param kmeans_eps: Tolerance of termination criterion of the k-means clustering used in hierarchical k-means tree building procedure, defaults to 0.0001
        :type kmeans_eps: float, optional
        :param kmeans_balanced: Use balanced k-means clustering, defaults to True
        :type kmeans_balanced: bool, optional
        :param hash: Hash features to a space of given size, value of this argument is saved with model weights, if None or 0 disable hashing, defaults to None
        :type hash: int, optional
        :param features_threshold: Prune features below given threshold, value of this argument is saved with model weights, defaults to 0
        :type features_threshold: float, optional
        :param norm: Unit norm feature vector, value of this argument is saved with model weights, defaults to True
        :type norm: bool, optional
        :param bias: Value of the bias features, value of this argument is saved with model weights, defaults to 1.0
        :type bias: float, optional
        :param optimizer: Optimizer used for training node classifiers {``'liblinear'``, ``'sgd'``, ``'adagrad'``}, defaults to ``'liblinear'``
        :type optimizer: str, optional
        :param loss: Loss optimized while training node classifiers {``'log'`` (alias ``'logistic'``), ``'l2'`` (alias ``'squaredHinge'``)}, defaults to ``'log'``
        :type loss: str, optional
        :param weights_threshold: Threshold value for pruning weights, defaults to 0.1
        :type weights_threshold: float, optional
        :param liblinear_c: LIBLINEAR cost co-efficient, inverse regularization strength, smaller values specify stronger regularization, makes effect only if ``optimizer='liblinear'``, defaults to 10.0
        :type liblinear_c: float, optional
        :param liblinear_eps: LIBLINEAR tolerance of termination criterion, makes effect only if ``optimizer='liblinear'``, defaults to 0.1
        :type liblinear_eps: float, optional
        :param liblinear_solver: Override LIBLINEAR solver set by loss parameter (default for ``loss='log'``: ``'L2R_LR_DUAL'``, for ``loss='l2'``: ``'L2R_L2LOSS_SVC_DUAL'``), makes effect only if ``optimizer='liblinear'``.
            Available solvers:

            - ``'L2R_LR_DUAL'``
            - ``'L2R_LR'``
            - ``'L1R_LR'``
            - ``'L2R_L2LOSS_SVC_DUAL'``
            - ``'L2R_L2LOSS_SVC'``
            - ``'L2R_L1LOSS_SVC_DUAL'``
            - ``'L1R_L2LOSS_SVC'``

            ``L2R_LR_DUAL`` and ``L2R_L2LOSS_SVC_DUAL`` usually work the best in XC setting, defaults to None
        :type liblinear_solver: str, optional
        :param liblinear_max_iter: Limits number of iteration by LIBLINEAR, makes effect only if ``optimizer='liblinear'``, defaults to 100
        :type liblinear_max_iter: int, optional
        :param eta: Step size (learning rate) for online optimizers, defaults to 1.0
        :type eta: float, optional
        :param epochs: Number of training epochs for online optimizers, defaults to 1
        :type epochs: int, optional
        :param adagrad_eps: Defines starting step size for AdaGrad, defaults to 0.001
        :type adagrad_eps: float, optional
        :param ensemble: Number of trees in the ensemble, defaults to 1
        :type ensemble: int, optional
        :param seed: Seed, If None use current system time, defaults to None
        :type seed: int, optional
        :param threads: Number of threads used for training and prediction, if 0 use number of available CPUs, if -1 use number of available CPUs - 1, defaults to 0
        :type threads: int, optional
        :param verbose: If True print progress, defaults to False
        :type verbose: bool, optional
        """
        all_params = Model._get_init_params(locals())
        all_params.update({"model": "hsm"})
        super(HSM, self).__init__(**all_params)


class BR(Model):
    """
    Binary Relevance (multi-label) classifier with linear estimators, using CPP core
    """

    def __init__(self,
                 output,

                 # Features params
                 hash=None,
                 features_threshold=0,
                 norm=True,
                 bias=1.0,

                 # Base classifiers params
                 optimizer='liblinear',
                 loss='log',
                 weights_threshold=0.1,
                 liblinear_c=10,
                 liblinear_eps=0.1,
                 liblinear_solver=None,
                 liblinear_max_iter=100,
                 eta=1.0,
                 epochs=1,
                 adagrad_eps=0.001,

                 # Prediction
                 load_as='map',

                 # Other
                 threads=0,
                 mem_limit=0,
                 verbose=0,
                 **kwargs):
        """
        Construct a Binary Relevance model.

        :param output: Directory where the model will be stored
        :type output: str
        :param hash: Hash features to a space of given size, value of this argument is saved with model weights, if None or 0 disable hashing, defaults to None
        :type hash: int, optional
        :param features_threshold: Prune features below given threshold, value of this argument is saved with model weights, defaults to 0
        :type features_threshold: float, optional
        :param norm: Unit norm feature vector, value of this argument is saved with model weights, defaults to True
        :type norm: bool, optional
        :param bias: Value of the bias features, value of this argument is saved with model weights, defaults to 1.0
        :type bias: float, optional
        :param optimizer: Optimizer used for training node classifiers {``'liblinear'``, ``'sgd'``, ``'adagrad'``}, defaults to ``'liblinear'``
        :type optimizer: str, optional
        :param loss: Loss optimized while training node classifiers {``'log'`` (alias ``'logistic'``), ``'l2'`` (alias ``'squaredHinge'``)}, defaults to ``'log'``
        :type loss: str, optional
        :param weights_threshold: Threshold value for pruning weights, defaults to 0.1
        :type weights_threshold: float, optional
        :param liblinear_c: LIBLINEAR cost co-efficient, inverse regularization strength, smaller values specify stronger regularization, makes effect only if ``optimizer='liblinear'``, defaults to 10.0
        :type liblinear_c: float, optional
        :param liblinear_eps: LIBLINEAR tolerance of termination criterion, makes effect only if ``optimizer='liblinear'``, defaults to 0.1
        :type liblinear_eps: float, optional
        :param liblinear_solver: Override LIBLINEAR solver set by loss parameter (default for ``loss='log'``: ``'L2R_LR_DUAL'``, for ``loss='l2'``: ``'L2R_L2LOSS_SVC_DUAL'``), makes effect only if ``optimizer='liblinear'``.
            Available solvers:

            - ``'L2R_LR_DUAL'``
            - ``'L2R_LR'``
            - ``'L1R_LR'``
            - ``'L2R_L2LOSS_SVC_DUAL'``
            - ``'L2R_L2LOSS_SVC'``
            - ``'L2R_L1LOSS_SVC_DUAL'``
            - ``'L1R_L2LOSS_SVC'``

            ``L2R_LR_DUAL`` and ``L2R_L2LOSS_SVC_DUAL`` usually work the best in XC setting, defaults to None
        :type liblinear_solver: str, optional
        :param liblinear_max_iter: Limits number of iteration by LIBLINEAR, makes effect only if ``optimizer='liblinear'``, defaults to 100
        :type liblinear_max_iter: int, optional
        :param eta: Step size (learning rate) for online optimizers, defaults to 1.0
        :type eta: float, optional
        :param epochs: Number of training epochs for online optimizers, defaults to 1
        :type epochs: int, optional
        :param adagrad_eps: Defines starting step size for AdaGrad, defaults to 0.001
        :type adagrad_eps: float, optional
        :param threads: Number of threads used for training and prediction, if 0 use number of available CPUs, if -1 use number of available CPUs - 1, defaults to 0
        :type threads: int, optional
        :param mem_limit: Maximum amount of memory (in G) available for training, if 0 use amount of available memory, defaults to 0
        :type mem_limit: float
        :param verbose: If True print progress, defaults to False
        :type verbose: bool, optional
        """
        all_params = Model._get_init_params(locals())
        all_params.update({"model": "br"})
        super(BR, self).__init__(**all_params)


class OVR(BR):
    """
    One Versus Rest (multi-class) classifier with linear estimators, using CPP core.
    """

    def __init__(self,
                 output,

                 # Features params
                 hash=None,
                 features_threshold=0,
                 norm=True,
                 bias=1.0,
                 pick_one_label_weighting=False,

                 # Base classifiers params
                 optimizer='liblinear',
                 loss='log',
                 weights_threshold=0.1,
                 liblinear_c=10,
                 liblinear_eps=0.1,
                 liblinear_solver=None,
                 liblinear_max_iter=100,
                 eta=1.0,
                 epochs=1,
                 adagrad_eps=0.001,

                 # Prediction
                 load_as='map',

                 # Other
                 threads=0,
                 mem_limit=0,
                 verbose=0,
                 **kwargs):
        """
        Construct a Multi-class One Versus Rest model.

        :param output: Directory where the model will be stored
        :type output: str
        :param hash: Hash features to a space of given size, value of this argument is saved with model weights, if None or 0 disable hashing, defaults to None
        :type hash: int, optional
        :param features_threshold: Prune features below given threshold, value of this argument is saved with model weights, defaults to 0
        :type features_threshold: float, optional
        :param norm: Unit norm feature vector, value of this argument is saved with model weights, defaults to True
        :type norm: bool, optional
        :param bias: Value of the bias features, value of this argument is saved with model weights, defaults to 1.0
        :type bias: float, optional
        :param pick_one_label_weighting: Allows to use multi-label data by transforming it into multi-class, defaults to False
        :type pick_one_label_weighting: bool, optional
        :param optimizer: Optimizer used for training node classifiers {``'liblinear'``, ``'sgd'``, ``'adagrad'``}, defaults to ``'liblinear'``
        :type optimizer: str, optional
        :param loss: Loss optimized while training node classifiers {``'log'`` (alias ``'logistic'``), ``'l2'`` (alias ``'squaredHinge'``)}, defaults to ``'log'``
        :type loss: str, optional
        :param weights_threshold: Threshold value for pruning weights, defaults to 0.1
        :type weights_threshold: float, optional
        :param liblinear_c: LIBLINEAR cost co-efficient, inverse regularization strength, smaller values specify stronger regularization, makes effect only if ``optimizer='liblinear'``, defaults to 10.0
        :type liblinear_c: float, optional
        :param liblinear_eps: LIBLINEAR tolerance of termination criterion, makes effect only if ``optimizer='liblinear'``, defaults to 0.1
        :type liblinear_eps: float, optional
        :param liblinear_solver: Override LIBLINEAR solver set by loss parameter (default for ``loss='log'``: ``'L2R_LR_DUAL'``, for ``loss='l2'``: ``'L2R_L2LOSS_SVC_DUAL'``), makes effect only if ``optimizer='liblinear'``.
            Available solvers:

            - ``'L2R_LR_DUAL'``
            - ``'L2R_LR'``
            - ``'L1R_LR'``
            - ``'L2R_L2LOSS_SVC_DUAL'``
            - ``'L2R_L2LOSS_SVC'``
            - ``'L2R_L1LOSS_SVC_DUAL'``
            - ``'L1R_L2LOSS_SVC'``

            ``L2R_LR_DUAL`` and ``L2R_L2LOSS_SVC_DUAL`` usually work the best in XC setting, defaults to None
        :type liblinear_solver: str, optional
        :param liblinear_max_iter: Limits number of iteration by LIBLINEAR, makes effect only if ``optimizer='liblinear'``, defaults to 100
        :type liblinear_max_iter: int, optional
        :param eta: Step size (learning rate) for online optimizers, defaults to 1.0
        :type eta: float, optional
        :param epochs: Number of training epochs for online optimizers, defaults to 1
        :type epochs: int, optional
        :param adagrad_eps: Defines starting step size for AdaGrad, defaults to 0.001
        :type adagrad_eps: float, optional
        :param threads: Number of threads used for training and prediction, if 0 use number of available CPUs, if -1 use number of available CPUs - 1, defaults to 0
        :type threads: int, optional
        :param mem_limit: Maximum amount of memory (in G) available for training, if 0 use amount of available memory, defaults to 0
        :type mem_limit: float
        :param verbose: If True print progress, defaults to False
        :type verbose: bool, optional
        """
        all_params = Model._get_init_params(locals())
        all_params.update({"model": "ovr"})
        super(OVR, self).__init__(**all_params)



import shutil
from napkinxc.datasets import load_dataset
from napkinxc.models import *
from napkinxc.measures import precision_at_k

from conf import *
MODEL_PATH = get_model_path(__file__)


def _test_model(model_class, model_config):
    X_train, Y_train = load_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)
    X_test, Y_test = load_dataset(TEST_DATASET, "test", root=TEST_DATA_PATH)

    model = model_class(MODEL_PATH, seed=TEST_SEED, **model_config)
    model.fit(X_train, Y_train)

    Y_pred = model.predict(X_test, top_k=1)
    p_at_1 = precision_at_k(Y_test, Y_pred, k=1)

    assert SCORE_RANGE[0] < p_at_1 < SCORE_RANGE[1]

    shutil.rmtree(MODEL_PATH, ignore_errors=True)


def test_plt_train_test():
    _test_model(PLT, {})


def test_hsm_train_test():
    _test_model(HSM, {"pick_one_label_weighting": True})


def test_br_train_test():
    _test_model(BR, {"optimizer": "adagrad", "epochs": 1})


def test_ovr_train_test():
    _test_model(OVR, {"pick_one_label_weighting": True})



import shutil
from time import time
import pytest
import numpy as np
from scipy.sparse import csr_matrix

from napkinxc.datasets import load_dataset, to_csr_matrix, to_np_matrix
from napkinxc.models import HSM, PLT
from napkinxc.measures import precision_at_k

from conf import *
MODEL_PATH = get_model_path(__file__)


def print_data_info(X):
    if not isinstance(X, list):
        print(f"Type (shape, dtype): {type(X)} ({X.shape}, {X.dtype})")
        if isinstance(X, csr_matrix):
            print(f"Internal types: indptr={type(X.indptr.dtype)}, indices={type(X.indices.dtype)}, data={type(X.data.dtype)}")
    else:
        print(f"Type (len): {type(X)} ({len(X)}), {X[0]}, X[0] type: {type(X[0])})")  #, X[0][0] type: {type(X[0][0])}")


def eval_data_for_dataset(X_train, Y_train, model_class):
    print_data_info(X_train)
    print_data_info(Y_train)
        
    model = model_class(MODEL_PATH, optimizer="adagrad", epochs=1, tree_type="huffman")

    t_start = time()
    model.fit(X_train, Y_train)
    print(f"Train time: {time() - t_start}")

    X_test, Y_test = load_dataset(TEST_DATASET, "test", root=TEST_DATA_PATH)

    t_start = time()
    Y_pred = model.predict(X_test, top_k=1)
    print(f"Predict time: {time() - t_start}")

    p_at_1 = precision_at_k(Y_test, Y_pred, k=1)
    print(f"P@1: {p_at_1}")
    assert 0.25 < p_at_1 < SCORE_RANGE[1]

    shutil.rmtree(MODEL_PATH, ignore_errors=True)


def test_multilabel_list_input():
    X_train, Y_train = load_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)
    eval_data_for_dataset(X_train, Y_train, PLT)


def test_multiclass_list_input():
    X_train, Y_train = load_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)

    X_train = X_train
    Y_train = [y[0] if len(y) else 0 for y in Y_train]
    eval_data_for_dataset(X_train, Y_train, HSM)


def test_multilabel_numpy_input():
    X_train, Y_train = load_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)

    #X_train = X_train.toarray()
    X_train = to_np_matrix(X_train)
    Y_train = to_np_matrix(Y_train)
    eval_data_for_dataset(X_train, Y_train, PLT)


def test_multilabel_numpy_float64_input():
    X_train, Y_train = load_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)

    #X_train = X_train.toarray()
    X_train = to_np_matrix(X_train, dtype=np.float64)
    Y_train = to_np_matrix(Y_train, dtype=np.float64)
    eval_data_for_dataset(X_train, Y_train, PLT)


def test_multiclass_numpy_float64_input():
    X_train, Y_train = load_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)

    #X_train = X_train.toarray()
    X_train = to_np_matrix(X_train, dtype=np.float64)
    Y_train = np.array([y[0] if len(y) else 0 for y in Y_train], dtype=np.float64)
    eval_data_for_dataset(X_train, Y_train, HSM)


def test_multiclass_numpy_input():
    X_train, Y_train = load_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)

    X_train = to_np_matrix(X_train, dtype=np.float64)
    Y_train = np.array([y[0] if len(y) else 0 for y in Y_train], dtype=np.int32)
    eval_data_for_dataset(X_train, Y_train, HSM)


def test_multiclass_numpy_int64_input():
    X_train, Y_train = load_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)

    X_train = to_np_matrix(X_train, dtype=np.float64)
    Y_train = np.array([y[0] if len(y) else 0 for y in Y_train], dtype=np.int64)
    eval_data_for_dataset(X_train, Y_train, HSM)



def test_multilabel_csr_input():
    X_train, Y_train = load_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)

    Y_train = to_csr_matrix(Y_train)
    eval_data_for_dataset(X_train, Y_train, PLT)


def test_multilabel_csr_input():
    X_train, Y_train = load_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)

    Y_train = to_csr_matrix(Y_train)
    eval_data_for_dataset(X_train, Y_train, PLT)


def test_multilabel_csr_float64_input():
    X_train, Y_train = load_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)

    X_train = to_csr_matrix(X_train, dtype=np.float64)
    Y_train = to_csr_matrix(Y_train)
    eval_data_for_dataset(X_train, Y_train, PLT)


def test_numpy_3d_input():
    size = 100
    X_train = np.ones((size, size, size))
    Y_train = np.ones((size))

    print(f"Type (shape, dtype): {type(X_train)} ({X_train.shape}, {X_train.dtype})")
    print(f"Type (shape, dtype): {type(Y_train)} ({Y_train.shape}, {Y_train.dtype})")

    with pytest.raises(ValueError):
        model = PLT(MODEL_PATH, optimizer="adagrad", epochs=1)
        model.fit(X_train, Y_train)



import shutil
import os
from napkinxc.datasets import load_dataset
from napkinxc.models import BR, PLT
from napkinxc.measures import precision_at_k

from conf import *
MODEL_PATH = get_model_path(__file__)


repeat = 5
model_configs = [
    {"tree_search_type": "exact"},
    {"tree_search_type": "beam"},
    {"ensemble": 3, "tree_search_type": "exact"},
    {"ensemble": 3, "tree_search_type": "beam"},
]
representation_configs = [
    {"load_as": "sparse"},
    {"load_as": "map"},
    {"load_as": "dense"},
]


def test_plt_exact_prediction_reproducibility():
    X_train, Y_train = load_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)
    X_test, Y_test = load_dataset(TEST_DATASET, "test", root=TEST_DATA_PATH)

    print("\n")
    for mc in model_configs:
        print("model config: ", mc)
        plt = PLT(MODEL_PATH, **mc)
        plt.fit(X_train, Y_train)
        Y_pred = plt.predict(X_test, top_k=1)
        p_at_1 = precision_at_k(Y_test, Y_pred, k=1)

        for rc in representation_configs:
            print("  prediction config: ", rc)
            for _ in range(repeat):
                plt = PLT(MODEL_PATH, **mc, **rc)
                Y_pred = plt.predict(X_test, top_k=1)
                assert p_at_1 == precision_at_k(Y_test, Y_pred, k=1)

        shutil.rmtree(MODEL_PATH, ignore_errors=True)


def test_seed_reproducibility():
    X_train, Y_train = load_dataset(TEST_DATASET, "train", root=TEST_DATA_PATH)
    X_test, Y_test = load_dataset(TEST_DATASET, "test", root=TEST_DATA_PATH)

    for i in range(repeat):
        plt_1 = PLT(MODEL_PATH + "-1", optimizer="adagrad", epochs=1, loss="log", seed=i)
        plt_1.fit(X_train, Y_train)
        Y_pred_1 = plt_1.predict(X_test, top_k=1)
        p_at_1_1 = precision_at_k(Y_test, Y_pred_1, k=1)
        tree_structure_1 = plt_1.get_tree_structure()

        plt_2 = PLT(MODEL_PATH + "-2", optimizer="adagrad", epochs=1, loss="log", seed=i)
        plt_2.fit(X_train, Y_train)
        Y_pred_2 = plt_2.predict(X_test, top_k=1)
        p_at_1_2 = precision_at_k(Y_test, Y_pred_2, k=1)
        tree_structure_2 = plt_2.get_tree_structure()

        assert len(set(tree_structure_1) - set(tree_structure_2)) == 0
        assert p_at_1_1 == p_at_1_2

        shutil.rmtree(MODEL_PATH + "-1", ignore_errors=True)
        shutil.rmtree(MODEL_PATH + "-2", ignore_errors=True)
